#ifndef ePaper_h
#define ePaper_h

// ceptimus 2017-12-14
// routines for driving E-paper displays
// lots copied from the example code on the Waveshare Wiki website https://www.waveshare.com/wiki/1.54inch_e-Paper_Module
// tested with 1.54-inch 200px x 200px unit but should work with others
// used 8MHz 3.3V Pro Mini for tests.  Beware! If you use a 5V model you probably need to use a logic level converter in the connection to the E-paper display - check the documentation for your display.
// display has 5000-byte write-only screen area and Arduino 328mega chips only 2 k of RAM so these routines use a 23LC1024 8-pin-DIL SPI serial RAM chip for paged memory.
// the 23LC1024 has 128 kbytes of RAM, is pretty cheap and works fast.  See http://ceptimus.co.uk/?p=166
// there are smaller capacity chips in the same family that are slightly cheaper.  You only really need 8 kbytes for the 200 x 200 display.

// ceptimus 2015-12-05
// fast, small, and simple interrupt-driven routines to use the 128 kbyte SPI RAM chip 23LC1024 on a Uno, Nano, Micro or other AtMega328 chip
// code and connections will need tweaking for use with other Arduino family members
// connections chip pin number - Arduino reference:
// 1 - D10, 2 - D12, 3 - VCC, 4 - Gnd, 5 - D11, 6 - D13, 7 - 5V, 8 - VCC
// if you want to be correct you can put 100K or so pull-up-resistors to pins 3 and 7 instead of wiring them direct to VCC - but it doesn't really matter.
// try to keep the wiring fairly short and neat.  If you must have longer wires, it helps to have a bypass capacitor 0.1uF connected close to the chip across the supply (pins 4 and 8).

// you don't have to use pin D10 for SS in SPI master mode - but if you use a different pin, then you must still configure D10 as an output (and use it freely without affecting SPI)
// don't leave D10 as an input (or if you do make sure it is kept pulled high).
// You might have more than one SPI ram chip, - then you need extra SS outputs for each chip but you can still use D10 for one of them.
// this code then needs tweaking to enable the appropriate chip at read/write time
// if you do that, you probably won't have to bother about the set-up part as sequential mode is the SPI RAM chip default anyway - I only put the set-up part in for completeness. 

// reference wiring
// E-paper  RAM chip   Arduino
// 3.3V     3, 7, 8    3.3V
// GND      4          GND
// DIN      5          D11 (MOSI)
// CLK      6          D13 (SCK)
// CS                  D6
// DC                  D9
// RST                 D8
// BUSY                D7
//          1          D10 Note: if you use a different chip select pin for the serial RAM chip, remember D10 must still be an output or an input tied high for SPI RAM access to work
//          2          D12 (MISO)

// although the serial RAM holds the full buffer we need to copy a 'letterbox' of serial RAM to and from ordinary Arduino RAM to work on it.
// 400 bytes is enough to hold 16 rasters of display ram 200px wide.  More works faster but 328Mega only has 2048 bytes total

#include <Arduino.h>

#define RAM_BUFFER_SIZE 400
#define LETTERBOX_HEIGHT ((RAM_BUFFER_SIZE) / (EPD_WIDTH >> 3))

// Pin definition - note CS_PIN moved from D10 in Waveshare example to D6 to allow the serial RAM chip to use D10
#define CS_PIN   6
#define BUSY_PIN 7
#define RST_PIN  8
#define DC_PIN   9

#define SPI_SS 10

// SPI clock speed for E-paper based on AtMega clock (F_CPU)
// E-paper has maximum clock speed of 4 MHz.  We want F_CPU / 2 for F_CPU <= 8MHz, then F_CPU / 4 for F_CPU <= 16MHz, and F_CPU / 8 for higher speeds (AtMega rated up to 20MHz)
#if F_CPU <= 16000000
  #define SPR1SPR0 0x00
  #if F_CPU <= 8000000
    #define SPI2XBIT 0x01
  #else
    #define SPI2XBIT 0x00
  #endif
#else
  #define SPR1SPR0 0x01
  #define SPI2XBIT 0x01
#endif

// Display resolution
#define EPD_WIDTH       200
#define EPD_HEIGHT      200

// EPD1IN54 commands
#define DRIVER_OUTPUT_CONTROL                       0x01
#define BOOSTER_SOFT_START_CONTROL                  0x0C
#define GATE_SCAN_START_POSITION                    0x0F
#define DEEP_SLEEP_MODE                             0x10
#define DATA_ENTRY_MODE_SETTING                     0x11
#define SW_RESET                                    0x12
#define TEMPERATURE_SENSOR_CONTROL                  0x1A
#define MASTER_ACTIVATION                           0x20
#define DISPLAY_UPDATE_CONTROL_1                    0x21
#define DISPLAY_UPDATE_CONTROL_2                    0x22
#define WRITE_RAM                                   0x24
#define WRITE_VCOM_REGISTER                         0x2C
#define WRITE_LUT_REGISTER                          0x32
#define SET_DUMMY_LINE_PERIOD                       0x3A
#define SET_GATE_TIME                               0x3B
#define BORDER_WAVEFORM_CONTROL                     0x3C
#define SET_RAM_X_ADDRESS_START_END_POSITION        0x44
#define SET_RAM_Y_ADDRESS_START_END_POSITION        0x45
#define SET_RAM_X_ADDRESS_COUNTER                   0x4E
#define SET_RAM_Y_ADDRESS_COUNTER                   0x4F
#define TERMINATE_FRAME_READ_WRITE                  0xFF

class ePaper {
  public:
    static void start(void);
    static void clearBuffer(bool white); // clear - the sram frame buffer.  If white is true then background will be white and vice-versa
    static void displayFrame(void); // refresh the display from the sram buffer.
    
    static void pixel(int x, int y, bool white); // set pixel (x, y) to white if white true, else black
    static void line(int x1, int y1, int x2, int y2, bool white); // draw line.
    static void rectangle(int x1, int y1, int x2, int y2, bool white); // draw unfilled rectangle with horizontal and vertical lines (x1, y1) and (x2, y2) are coordinates of diagonally opposite corners.
    static void filledRectangle(int x1, int y1, int x2, int y2, bool white);
    static void circle(int centreX, int centreY, int radius, bool white); // unfilled circle
    static void filledCircle(int centreX, int centreY, int radius, bool white);
    static void textSmall(int x, int y, const char *s, bool white); // 6x8 font
    static void text(int x, int y, const char *s, bool white); // 6x8 font stretched to double-size 12x16
    static void sleep(void);
    static void wake(void);
    static void fullUpdateMode(void);
    static void partialUpdateMode(void);
  private:
    const static uint8_t PROGMEM font[];
    const static uint8_t PROGMEM stretch[]; // pixel doubling look-up for double-size 'large' characters
    const static uint8_t PROGMEM lut_full_update[];
    const static uint8_t PROGMEM lut_partial_update[];

    static uint8_t ramBuf[];
    static int letterboxTop;
    static int centreX;
    static int centreY;
    static bool white;
    static void circle(int centreX, int centreY, int radius, bool white, void callback(int x, int y));
    static void circleCallback(int x, int y); // set pixel (x + centreX, y + centreY) to white if white true, else black
    static void filledCircleCallback(int x, int y);
    static void firstQuadrant(int limit, uint32_t square, void callback(int x, int y));
    static void secondAndThirdQuadrants(int limit, uint32_t square, void callback(int x, int y));
    static void fourthQuadrant(int limit, uint32_t square, void callback(int x, int y));
    
    static void positionLetterbox(int top);
    static void displayChar6x8(int x, int y, char c, bool white);
    static void displayChar12x16(int x, int y, char c, bool white);
    static void setMemoryArea(int x_start, int y_start, int x_end, int y_end);
    static void setMemoryPointer(int x, int y);
    static void setFrameMemory(uint8_t *image_buffer, int x, int y, int image_width, int image_height);
    static void sendCommand(uint8_t command);
    static void sendData(uint8_t data);
    static void waitUntilIdle(void);
    static void setLutFullUpdate(bool full);
    static void spiTransfer(uint8_t data);

    static void initSpiDriver(void);
    static void writeSpiRam(uint8_t *src, uint32_t address, uint16_t count); // src is pointer to normal Arduino SRAM
    static void readSpiRam(uint8_t *dest, uint32_t address, uint16_t count); // dest is pointer to normal Arduino SRAM
    static void writeEpaperSpi(uint8_t data); // send one byte to e-Paper
    static void writeEpaperSpi(uint8_t *src, uint16_t count); // src is pointer to 1st byte in normal Arduino SRAM, count is number of bytes to transfer
    static void startTransmission(uint8_t *ptr, uint32_t address, uint16_t count, uint8_t opcode, uint8_t mode);    
};

#endif
